/*
   Copyright (C) 2008
   	Sergey S. Dukachyov  iam@dukess.ru


   The TheBatAttachmentPathDoctor is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2.0 of the License, or (at your option) any later version.


*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>       /* asctime, gmtime */
#include <string.h>     /* for strlen */
#include "crc32_jamcrc.h"
#include "getopt.h"
#include "base64.h"
#ifndef bool 
#define bool int
#endif
#ifndef true 
#define true 1
#endif
#ifndef false 
#define false 0
#endif
#pragma hdrstop

//---------------------------------------------------------------------------

struct t_mail {
    int id;
    unsigned char *rawheader;
    unsigned long size;
    unsigned char *content;
    unsigned long received_time;
    unsigned int colorgroup;
    bool is_deleted;
    bool was_read;
    bool was_answered;
    bool is_parked;
    bool has_attachement;
    bool has_deleted_attachement;
    bool is_flagged;
    bool was_forwarded;
    
    bool high_priority;
    bool low_priority;
};
static bool debug = false;
#define IS_DELETED               1
#define IS_READ                  2
#define WAS_ANSWERED             4
#define IS_PARKED                8
#define HAS_ATTACHEMENT         16
#define HAS_DELETED_ATTACHEMENT 32
#define IS_FLAGGED              64
#define WAS_FORWARDED          128

int tbbFileRepair(char *sourcefilename, char *resultfilename, char *newPathToAttach);

#pragma argsused
int main(int argc, char* argv[])
	{
//    
int c;
        bool auto_mode = false;
        char *FileName = NULL;
        char *FileName_old = NULL;
        char *PathToAttach = NULL;
        int ExitCode = 0;
// a -   ,    Inbox  Sent      Attach
// f -   
// p -     
// h - 
// d -  
        const char *args = "f:p:hda";

        /* Process command line options */
        while ((c = getopt(argc, argv, args)) != -1)
        {
                switch (c)
                {
                  case 'a':
                        auto_mode = true;
                        break;
                  case 'f':
                        if (optarg == NULL || *optarg == '\0')
                        {
                                (void) fprintf(stderr, "Illegal file name: %s\n",
                                               optarg);
                                exit(0);
                        }
                        FileName = malloc(strlen(optarg) + 1);
                        if(!FileName){return 1;}
                        memset(FileName, 0, strlen(optarg) + 1);
                        strcpy(FileName, optarg);
                        FileName_old = malloc(strlen(optarg) + 5);
                        if(!FileName_old){free(FileName); if(PathToAttach) free(PathToAttach);return 1;}
                        memset(FileName_old, 0, strlen(optarg) + 5);
                        sprintf(FileName_old, "%s.old", optarg);
                        break;
                  case 'p':
                        if (optarg == NULL || *optarg == '\0')
                        {
                                (void) fprintf(stderr, "Illegal file name: %s\n",
                                               optarg);
                                exit(0);
                        }
                        PathToAttach = malloc(strlen(optarg) + 1);
                        if(!PathToAttach) {free(PathToAttach); if(FileName) free(FileName); if(FileName_old) free(FileName_old); return 1;}
                        memset(PathToAttach, 0, strlen(optarg) + 1);
                        memcpy(PathToAttach, optarg, strlen(optarg));
                        break;
                  case 'h':
                        printf ("Usage: bad {-a path|-f filename -p path} [-d] [-h]\n");
                        printf ("Options:\n");
                        printf ("   -a path                   - path to The Bat! mailbox directory\n");
                        printf ("   -f                        - path and filename of The Bat! mail database (*.tbb file)\n");
                        printf ("   -d                        - debug mode\n");
                        printf ("   -h                        - this help\n");
                        exit(0);
                        break;
                  case 'd':
                        debug = true;
                        break;
                }
        }                     

 if (auto_mode != true && (!FileName || !PathToAttach))
  {
  printf ("Usage: bad {-a [path_to_mailbox]| -f path_to_messages.tbb -p path_to_attach_dir } [-d] [-h]\n");
  return ExitCode;
  }
  
 if (auto_mode == true)
  {
  printf ("Ooops! Auto mode doesn't work now.\n");
  return ExitCode;
  }

	if(!tbbFileRepair(FileName, "messages.tmp", PathToAttach))
          {
          rename(FileName, FileName_old);
          rename("messages.tmp", FileName);
          ExitCode = 1;
          }
        free(FileName_old);
        free(FileName);
        free(PathToAttach);
	return ExitCode;
	}

// swap bytes for Little Endian to Big Endian conversion long integer values
/*int lbe_convert (int bEnum)
  {
    int lEnum;
    char *lE = (char*) &lEnum;
    char *bE = (char*) &bEnum;
    lE[0] = bE[3];
    lE[1] = bE[2];
    lE[2] = bE[1];
    lE[3] = bE[0];
    return lEnum;
  }

*/

void digit_hexprint(unsigned char character)   {
    printf("%02x ", character);
}

void bitstream_hexprint(unsigned char *buffer, int length)   {
    int i;
    if (!debug) return;
    for (i = 0; i < length; i++)  {
        digit_hexprint(buffer[i]);
    }
    printf("\n");
}

bool bitstreams_are_equal(unsigned char *buffer1, int length_buffer1, unsigned char *buffer2, int length_buffer2)  {
    int i;
		if (length_buffer1 != length_buffer2)
				return false;

		for (i = 0; i < length_buffer1; i++)
        if (buffer1[i] != buffer2[i])
            return false;
    return true;
}

bool is_tbb_file(FILE *filehandle)   {
    int result = true;
		int length_of_magic_number = 4;
		unsigned char *first_bytes_from_file;
		int read_bytes;
		char *tbb_magic_number = (char*) malloc(length_of_magic_number * sizeof(char));

/*    sprintf(tbb_magic_number, "%c%c%c%c%c%c", 0x20, 0x06, 0x79, 0x19, 0x08, 0x0C);
*/
    sprintf(tbb_magic_number, "%c%c%c%c", 0x20, 0x06, 0x79, 0x19);
		first_bytes_from_file = (unsigned char*) malloc(length_of_magic_number * sizeof(char));

    fseek(filehandle, 0, SEEK_SET);
		read_bytes = fread(first_bytes_from_file, sizeof(char), length_of_magic_number, filehandle);
    if (! bitstreams_are_equal(first_bytes_from_file, read_bytes, tbb_magic_number, length_of_magic_number))
        result = false;
    free(tbb_magic_number);
    free(first_bytes_from_file);
    return result;
}


unsigned char* read_per_mail_header(FILE *filehandle) {
    int sizeof_permail_header = 48;
    unsigned char *header = malloc(sizeof_permail_header * sizeof(char));
    int read_bytes = fread (header, sizeof(char), sizeof_permail_header, filehandle);
    if (read_bytes == 0 && feof(filehandle))    {
        free(header);
        header = NULL;
        printf("end of file reached!\n");
    }
    else
        bitstream_hexprint(header, read_bytes);
    return header;
}

int is_mail_header(unsigned char *header)    {
    int result = true;
    int length_of_magic_number = 5;
    char *header_magic_number = (char*) malloc(length_of_magic_number * sizeof(char));
    sprintf(header_magic_number, "%c%c%c%c%c", 0x21, 0x9, 0x70, 0x19, 0x30);
    if (! bitstreams_are_equal(header, length_of_magic_number, (unsigned char*) header_magic_number, length_of_magic_number))
        result = false;
    free(header_magic_number);
    return result;

}

void init_header(struct t_mail *header) {
    header->id = 0;
    header->size = 0;
    header->received_time = 0;
    header->content = NULL;
    header->is_deleted = false;
    header->was_read = false;
    header->was_answered = false;
    header->is_parked = false;
    header->has_attachement = false;
    header->has_deleted_attachement = false;
    header->is_flagged = false;
    header->was_forwarded = false;
    header->high_priority = false;
    header->low_priority = false;
}

void parse_message_status_flag(struct t_mail *header, char flag) {
    if (flag & IS_DELETED)
        header->is_deleted = true;
    if (flag & IS_READ)
        header->was_read = true;
    if (flag & WAS_ANSWERED)
        header->was_answered = true;
    if (flag & IS_PARKED)
        header->is_parked = true;
    if (flag & HAS_ATTACHEMENT)
        header->has_attachement = true;
    if (flag & HAS_DELETED_ATTACHEMENT)
        header->has_deleted_attachement = true;
    if (flag & IS_FLAGGED)
        header->is_flagged = true;
    if (flag & WAS_FORWARDED)
        header->was_forwarded = true;
}

unsigned long convert_bitstream_to_number(unsigned char *bitstream) {
    int i;
    unsigned long result = 0;
    unsigned long factor = 1;
    for (i = 0; i < 4; i++) {
        result += factor * bitstream[i];
        factor *= 256;
    }
    return result;
}

void parse_mail_priority(struct t_mail *header, unsigned char* bitstream) {
    if (convert_bitstream_to_number(bitstream) == 5) {
        header->high_priority = true;
    }
    if (convert_bitstream_to_number(bitstream) == (((unsigned long) 1) << 32) - 5) {
        header->low_priority = true;
    }
}


struct t_mail* parse_mail_header(unsigned char *headerstring)    {
    struct t_mail *header = malloc(sizeof(struct t_mail));
    if (! is_mail_header(headerstring))   {
        printf("Your TBB file has a currently unknown mail header. Please contact the authors.\n");
        free(header);
        return NULL;
    }
    init_header(header);
    header->received_time = convert_bitstream_to_number(headerstring + 12);
    header->id = 256 * headerstring[17] + headerstring[16];
    parse_message_status_flag(header, headerstring[20]);
    header->colorgroup = convert_bitstream_to_number(headerstring + 28);
    parse_mail_priority(header, headerstring + 32);
    header->size = convert_bitstream_to_number(headerstring + 36);
    return header;
}

void print_mail_infos(struct t_mail *mail)  {
    printf("id is: %d\n", mail->id);
    printf("received: %lu\n", mail->received_time);
    printf("color group: %x\n", mail->colorgroup);
    printf("size is: %lu bytes\n", mail->size);
    if (mail->is_deleted)
        printf("  deleted\n");
    if (mail->was_read)
        printf("  read\n");
    if (mail->was_answered)
        printf("  answered\n");
    if (mail->is_parked)
        printf("  parked\n");
    if (mail->has_attachement)
        printf("  has attachements\n");
    if (mail->has_deleted_attachement)
        printf("  has deleted attachements\n");
    if (mail->is_flagged)
        printf("  flagged\n");
    if (mail->was_forwarded)
        printf("  forwarded\n");    
    if (mail->high_priority)
        printf("  high priority\n");    
    if (mail->low_priority)
        printf("  low priority\n");    
    printf("\n");
}

struct t_mail* extract_mail(FILE *filehandle) {
    unsigned char *header = read_per_mail_header(filehandle);
		struct t_mail *mail = NULL;
		int read_bytes;
    if (header != NULL) {
        mail = parse_mail_header(header);
        if (mail != NULL)   {
            if(debug) print_mail_infos(mail);
            mail->content = malloc(mail->size + 1);
            mail->rawheader = header;
	    read_bytes = fread(mail->content, sizeof(char), mail->size, filehandle);
// make null-terminated string
            mail->content[mail->size] = 0;
            if (read_bytes != mail->size) {
                printf("TBB file seems to be corrupt! Size of mail was supposed to be %lu bytes but the program could only read %d bytes!\n", mail->size, read_bytes);
                free(mail->content);
                free(mail);
                free(header);
                mail = NULL;
            }
        }
//        free(header);
    }
    return mail;
}

long int strstr_int(char *s1, char *s2)
  {
  int i1, i2, j;
  i1 = strlen(s1);
  i2 = strlen(s2);
  for (j = 0; j < i1; j++)
    {
    if(!memcmp(s1 + j, s2, i2))
      {
      return j;
      }
    }
  return -1;  
  }

int tbbFileRepair(char *sourcefilename, char *resultfilename, char *newPathToAttach)
	{
/*	char tbbFileHeaderMagicNumber[4] = ;
*/
	int tbbFileHeaderSize;
	struct t_mail *mail;
        char *memory = NULL;
        FILE *sourceTbbFileHandler, *resultTbbFileHandler;
 	sourceTbbFileHandler = fopen(sourcefilename, "rb");
        if(!sourceTbbFileHandler)
          {
          printf("File %s not found\n", sourcefilename);
          return 1;
          }
	resultTbbFileHandler = fopen(resultfilename, "wb");
        if(!resultTbbFileHandler)
          {
          printf("Can't create file %s\n", resultfilename);
          return 1;
          }
        if(is_tbb_file(sourceTbbFileHandler))
          {
	printf("It is a TBB file\n");
          }
        else
          {
	printf("It is not a TBB file\n");
          return 2;
          }

	fread(&tbbFileHeaderSize, sizeof(int) ,1, sourceTbbFileHandler);
	printf("Header size is %d\n", tbbFileHeaderSize);
//copy header to new file
        memory = malloc(tbbFileHeaderSize);
        if(!memory)
          {
          printf("Can't alloc %d bytes for header\n", tbbFileHeaderSize);
          return 4;
          }
	fseek(sourceTbbFileHandler, 0, SEEK_SET);
	fread(memory, sizeof(char) ,tbbFileHeaderSize, sourceTbbFileHandler);
	fwrite(memory, sizeof(char), tbbFileHeaderSize, resultTbbFileHandler);
        free(memory);
//	fseek(sourceTbbFileHandler, tbbFileHeaderSize, SEEK_SET);
//convert data

	mail = extract_mail(sourceTbbFileHandler);
        while (mail != NULL)
            {
          long int pos1 = 0;  // begin position of substring "X-Content-File:"
          long int pos2 = 0;  // \n symbols position
          int AttachCount = 0;
          char *Base64EncodedFilename;
          int i = 0;
          int temppos = 0;
// Count all attaches
          do
            {
            temppos += pos1;
            pos1 = strstr_int(mail->content + temppos, "X-Content-File:");
            temppos += 15;
            AttachCount++;
            }
          while (pos1 != -1);
          AttachCount--;
//if no attaches then fast copy to destination file...
          if(!AttachCount)
            {
            fwrite(mail->rawheader, sizeof(char), 48, resultTbbFileHandler);
	    fwrite(mail->content, sizeof(char), mail->size, resultTbbFileHandler);
            }
          else
            {
//...else convert paths and save to destination file.
            int temppos_src = 0, temppos_dst = 0, memcpy_size = 0;
            char *tempMailContent = malloc(mail->size + AttachCount * strlen(newPathToAttach) + 1);
            unsigned long CRC32 = 0;
            if (!tempMailContent) {printf("Fault! Can't alloc memory for tempMailContent\n"); return 4;}
            memset(tempMailContent, 0, mail->size + AttachCount * strlen(newPathToAttach) + 1);

            pos1 = strstr_int(mail->content, "X-Content-File: ");
            pos2 = strstr_int(mail->content + pos1 + 16, "\r\n");
            memcpy_size = pos1 + 16;
            while(pos1 != -1)
              {
              memcpy(tempMailContent + temppos_dst, mail->content + temppos_src, memcpy_size);
              temppos_src += pos1 + 16;
              temppos_dst += pos1 + 16;
//              memcpy(tempMailContent + temppos_dst, " ", 1);
//              temppos_dst += 1;

//Base64 encoded filename?
              Base64EncodedFilename = malloc(pos2 + 1);
              memset(Base64EncodedFilename, 0, pos2 + 1);
              memcpy(Base64EncodedFilename, mail->content + temppos_src, pos2);
               if(strstr(Base64EncodedFilename, "=?") && strstr(Base64EncodedFilename, "?B?"))
                {
                char *Base64DecodedFilename = malloc(pos2 + 1);
                char *Base64RecodedFilename = malloc((pos2 + 1 + strlen(newPathToAttach)) * 1.5);
                char *Base64PrerecodedFilename = malloc(pos2 + 1 + strlen(newPathToAttach));
        if(!Base64DecodedFilename)
          {
          printf("Can't alloc %d bytes for Base64DecodedFilename\n", tbbFileHeaderSize);
          return 4;
          }
        if(!Base64RecodedFilename)
          {
          printf("Can't alloc %d bytes for Base64RecodedFilename\n", tbbFileHeaderSize);
          return 4;
          }
        if(!Base64PrerecodedFilename)
          {
          printf("Can't alloc %d bytes for Base64PrerecodedFilename\n", tbbFileHeaderSize);
          return 4;
          }
                memset(Base64DecodedFilename, 0, pos2 + 1);
                memset(Base64RecodedFilename, 0, (pos2 + 1 + strlen(newPathToAttach)) * 1.5);
                memset(Base64PrerecodedFilename, 0, pos2 + 1 + strlen(newPathToAttach));

                memcpy(Base64RecodedFilename, Base64EncodedFilename, strstr_int(Base64EncodedFilename, "?="));
                strcpy(Base64EncodedFilename, strstr(Base64RecodedFilename, "?B?") + 3);
                memset(Base64RecodedFilename, 0, (pos2 + 1 + strlen(newPathToAttach)) * 1.5); //nulling

//decode filename
//                base64_decode(strstr(Base64EncodedFilename, "?B?") + 3, Base64DecodedFilename, pos2);
                base64_decode(Base64EncodedFilename, Base64DecodedFilename, pos2);

//ANSI to OEM convert Win1251 -> CP866
/*                for(i = 0; i < strlen(Base64DecodedFilename); i++)
                  {
                  unsigned char Temp = (unsigned char)Base64DecodedFilename[i];
                  if(Temp > 127 + 64 && Temp < 176 + 64) Base64DecodedFilename[i] -= 64;
                  else if(Temp > 223 + 16)Base64DecodedFilename[i] -= 16;
                  else if(Temp == 168)Base64DecodedFilename[i] = 240;
                  else if(Temp == 184)Base64DecodedFilename[i] = 241;
                  }
*/
// search last backslash in file path
                for (i = pos2 - 1; i >= 0; i--)
                  {
                  if (Base64DecodedFilename[i] == '\\') break;
                  }
//                printf("%s|%s\n", Base64EncodedFilename , Base64DecodedFilename);
                strcat(Base64PrerecodedFilename, newPathToAttach);
                strcat(Base64PrerecodedFilename, Base64DecodedFilename + i);

if(false)
  {
//not encode filename
//                base64_encode(Base64PrerecodedFilename, strlen(Base64PrerecodedFilename), Base64RecodedFilename, (pos2 + 1 + strlen(newPathToAttach)) * 1.5);
                memcpy(tempMailContent + temppos_dst, Base64PrerecodedFilename, strlen(Base64PrerecodedFilename));
                temppos_dst += strlen(Base64PrerecodedFilename);
  }
else
  {
//OEM to ANSI convert CP866 -> Win1251
/*                for(i = 0; i < strlen(Base64PrerecodedFilename); i++)
                  {
                  unsigned char Temp = (unsigned char)Base64PrerecodedFilename[i];
                  if(Temp > 127 && Temp < 176) Base64PrerecodedFilename[i] += 64;
                  else if(Temp > 223 && Temp < 240)Base64PrerecodedFilename[i] += 16;
                  else if(Temp == 240)Base64PrerecodedFilename[i] = 168;
                  else if(Temp == 241)Base64PrerecodedFilename[i] = 184;
                  }
*/
//encode filename
                base64_encode(Base64PrerecodedFilename, strlen(Base64PrerecodedFilename), Base64RecodedFilename, (pos2 + 1 + strlen(newPathToAttach)) * 1.5);
                memcpy(tempMailContent + temppos_dst, mail->content + temppos_src, pos2);
                temppos_dst += strstr_int(tempMailContent + temppos_dst, "?B?") + 3;
                memcpy(tempMailContent + temppos_dst, Base64RecodedFilename, strlen(Base64RecodedFilename));
                temppos_dst += strlen(Base64RecodedFilename);
                memcpy(tempMailContent + temppos_dst, "?=", 2);
                temppos_dst += 2;
   }
//
                temppos_src += pos2;
                free(Base64DecodedFilename);
                free(Base64RecodedFilename);
                free(Base64PrerecodedFilename);
                }
              else
                {
                memcpy(tempMailContent + temppos_dst, newPathToAttach, strlen(newPathToAttach));
                temppos_dst += strlen(newPathToAttach);

// search last backslash in file path
                for (i = temppos_src + pos2 - 1; i >= temppos_src; i--)
                  {
//                  printf("%c", mail->content[i]);
                  if (mail->content[i] == '\\') {temppos_src = i; break;}
                  }
                }

              pos1 = strstr_int(mail->content + temppos_src, "X-Content-File: ");
              pos2 = strstr_int(mail->content + temppos_src + pos1 + 16, "\r\n");
              memcpy_size = pos1 + 16;
              free(Base64EncodedFilename);
              }
            memcpy(tempMailContent + temppos_dst, mail->content + temppos_src, strlen(mail->content + temppos_src));
//            if(debug) printf("Data:\n%s\n", tempMailContent);
            if(debug) printf("Size: old - %d, new - %d\n", mail->size, strlen (tempMailContent));
            mail->size = strlen (tempMailContent);
            CRC32 = Crc32_jamcrcCalc (0xFFFFFFFF, tempMailContent);  // crc = 0xFFFFFFFF
//            CRC32 = Crc32_jamcrcCalc_bin (0xFFFFFFFF, tempMailContent, strlen (tempMailContent));  // crc = 0xFFFFFFFF
            if(debug) printf("Old CRC32:%lX   New CRC32:%lX\n", ((unsigned long *)(mail->rawheader + 8))[0], CRC32);
            ((unsigned long *)(mail->rawheader + 8))[0] = CRC32;
            ((int *)(mail->rawheader + 36))[0] = mail->size;
            fwrite(mail->rawheader, sizeof(char), 48, resultTbbFileHandler);
	    fwrite(tempMailContent, sizeof(char), mail->size, resultTbbFileHandler);
            free(tempMailContent);
            }
          free(mail->rawheader);
          free(mail->content);
          free(mail);
          mail = extract_mail(sourceTbbFileHandler);
          }

	fclose(sourceTbbFileHandler);
        fclose(resultTbbFileHandler);
        return 0;
	};